from drf_yasg.inspectors import JSONFieldInspector


class ValidatedJSONFieldInspector(JSONFieldInspector):

    def add_manual_fields(self, serializer_or_field, schema):
        """Set fields from the ``swagger_schem_fields`` attribute on the Meta class. This method is called
        only for serializers or fields that are converted into ``openapi.Schema`` objects.
        :param serializer_or_field: serializer or field instance
        :param openapi.Schema schema: the schema object to be modified in-place
        """

        coreapi_schema = getattr(serializer_or_field, 'coreapi_schema', {})
        if coreapi_schema:
            for attr, val in coreapi_schema.items():
                setattr(schema, attr, val)

        meta = getattr(serializer_or_field, 'Meta', None)
        swagger_schema_fields = getattr(meta, 'swagger_schema_fields', {})
        if swagger_schema_fields:
            for attr, val in swagger_schema_fields.items():
                setattr(schema, attr, val)




DEFAULT_FIELD_INSPECTORS = [
    'django_validated_jsonfield.inspectors.ValidatedJSONFieldInspector',
    'drf_yasg.inspectors.CamelCaseJSONFilter',
    'drf_yasg.inspectors.RecursiveFieldInspector',
    'drf_yasg.inspectors.ReferencingSerializerInspector',
    'drf_yasg.inspectors.ChoiceFieldInspector',
    'drf_yasg.inspectors.FileFieldInspector',
    'drf_yasg.inspectors.DictFieldInspector',
    'drf_yasg.inspectors.JSONFieldInspector',
    'drf_yasg.inspectors.HiddenFieldInspector',
    'drf_yasg.inspectors.RelatedFieldInspector',
    'drf_yasg.inspectors.SerializerMethodFieldInspector',
    'drf_yasg.inspectors.SimpleFieldInspector',
    'drf_yasg.inspectors.StringDefaultFieldInspector',
]




import copy
import json
from openapi_schema_to_json_schema import to_json_schema
import jsonschema

try:
    from django.db.models import JSONField
except:
    from django.contrib.postgres.fields import JSONField

from django.contrib.staticfiles import finders
from django.core import checks, exceptions
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.functional import cached_property

#########################################################


def extend_with_default(validator_class):
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(validator, properties, instance, schema):
        for property_, subschema in properties.items():
            if "default" in subschema and not isinstance(instance, list):
                instance.setdefault(property_, subschema["default"])

        for error in validate_properties(
            validator, properties, instance, schema,
        ):
            yield error

    return jsonschema.validators.extend(
        validator_class, {"properties": set_defaults},
    )
ExtendedJsonValidatorWithDefault = extend_with_default(jsonschema.Draft7Validator)


class ValidatedJSONField(JSONField):
    """
    A models.JSONField subclass that supports the JSON schema validation.
    """

    @property
    def _get_default(self):
        try:
            obj = [] if self.isArray else {}
            if(self.json_validator_cls):
                self.json_validator_cls.validate(obj)
        except:
            obj = None
        return lambda: obj

    def run_validators(self, value):
        super().run_validators(value)

        if(self.json_validator_cls):
            errors = [ValidationError({str(error.path):error.message}) for error in self.json_validator_cls.iter_errors(value)]
            if(errors):
                raise ValidationError(errors)

    def __init__(self, *args, schema=None, default=None, blank=False, **kwargs):
        self.schema = schema
        self.jsschema = schema
        if schema is not None:
            self.jsschema = to_json_schema(schema, {"supportPatternProperties": True})
            self.json_validator_cls = ExtendedJsonValidatorWithDefault(self.jsschema)
        else:
            self.json_validator_cls = None

        self.isArray = False
        if(schema is not None):
            if(default==list or default==[] or str(schema.get("type","")).lower()=="array"):
                self.isArray=True
            elif(default==dict or default=={} or str(schema.get("type","")).lower()=="object"):
                self.isArray=False
            elif(default is not None):
                print("#WARNING# ValidatedJSONField: default argument is not allowed, as default is taken from schema", default)

            if(None and blank):
                print("#WARNING# ValidatedJSONField: blank argument can't be set to True otherwise validation is not run")

        super().__init__(*args, blank=False, default=None, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_validated_jsonfield']

package_data = \
{'': ['*']}

install_requires = \
['django>=2.2.0',
 'djangorestframework>=3.11.0',
 'drf_yasg',
 'jsonschema',
 'jsonschema-to-openapi>=0.2.1',
 'py-openapi-schema-to-json-schema>=0.0.3,<0.0.4']

setup_kwargs = {
    'name': 'django-validated-jsonfield',
    'version': '1.1.7',
    'description': 'Add a schema with validation to your jsonfield',
    'long_description': '# django_validated_jsonfield\n\ndjango_validated_jsonfield is an inplace replacement for django JSONField which is validated against a json_schema\n\n- A json_schema can be added to the field description\n- The json_schema is used to validate the jsonfield data during full_clean or during restframework validation\n- The field is also documented on Swagger (via drf_yasg specific inspectors)\n- Missing data in the json structure is initialized/filled to the schema\'s default values\n\n# Usage\n\n## import\n\nReplace your JSONField import\n\n```python\nfrom django.db.models import JSONField` or `from django.contrib.postgres.fields import JSONField\n```\n\nby \n\n```python\nfrom django_validated_jsonfield import ValidatedJSONField as JSONField\n```\n\n## models\n```python\nclass MyModel(models.Model):\n    _data_schema = {\n        "type": "object",\n        "properties": {\n            "en": {"type": ["string", "null"], "default":"",   "example":"Name", "title":"Name"}\n            "fr": {"type": ["string", "null"], "default":"",   "example":"Nom",  "title":"Nom"}\n        },\n        "default": {}, #note the top level default\n        "additionalProperties": False,\n    }\n\n    data = JSONField(schema=_data_schema, editable=True)\n```\n\n## Rest Framework serializers\n\n### As defined field:\n\nreplace\n```python\nfrom rest_framework.serializers import JSONField\n```\nby\n```python\nfrom django_validated_jsonfield import ValidatedJSONFieldSerializer as JSONField\n```\n\nThe updated JSONField expect to receive a "schema" and "json_validator_cls" argument at initilization.\n\n### In serializers.ModelSerializer:\n\nAdd the ValidatedJsonModelSerializerMixin to your ModelSerializer classes.\n\n```python\nfrom django_validated_jsonfield import ValidatedJsonModelSerializerMixin\n\nclass MyModelSerializer(ValidatedJsonModelSerializerMixin, serializers.ModelSerializer):\n    class Meta:\n        model = MyModel\n        exclude = []\n```\n\n\n## Swagger documentation (drf_yasg)\n\nadd the following block to your django settings.py file\n\n```python\nfrom django_validated_jsonfield.yasg import DEFAULT_FIELD_INSPECTORS\nSWAGGER_SETTINGS = {\n    \'DEFAULT_FIELD_INSPECTORS\': DEFAULT_FIELD_INSPECTORS\n}\n```\n\nyou should see the json field of your model being documented in Swagger\n\n\n# Additional remarks\n\n## json schema default\n\nif the default field is provided in the json schema, the data will be initialized to the default value (if missing).\nThe feature works well only parents of nested fields in the json schema have a default themselves (to list or dict)\n--> Note the "top level default" in the example above.\n\n',
    'author': 'Loic Quertenmont',
    'author_email': 'loic@youmeal.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/loic.quertenmont/django_validated_jsonfield',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import copy

from jsonschema_to_openapi.convert import convert as convert_to_openapi
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from openapi_schema_to_json_schema import to_json_schema

from .fields import ValidatedJSONField as ModelValidatedJSONField
from .fields import ExtendedJsonValidatorWithDefault

class ValidatedJSONField(serializers.JSONField):
    def __init__(self, *args, initial=None, validators=[], **kwargs):
        self.schema = kwargs.pop("schema", None)
        if(self.schema is not None):
            self.coreapi_schema = convert_to_openapi(copy.deepcopy(self.schema) )
        self.json_validator_cls = kwargs.pop("json_validator_cls", None)
        if(self.json_validator_cls is None and self.schema is not None):
            jsschema = to_json_schema(schema, {"supportPatternProperties": True})            
            self.json_validator_cls = ExtendedJsonValidatorWithDefault(jsschema)

        #MIXING validators among multiple ModelValidatedJSONField, we should try not to use the validators argument
        def json_validator(value):
            if(self.json_validator_cls):
                errors = {".".join([self.field_name]+[str(p) for p in error.path]):error.message for error in self.json_validator_cls.iter_errors(value)}
                if(errors):
                    raise ValidationError(errors)

        self.validators =list([json_validator]+validators)

        super().__init__(*args, **kwargs)





class ValidatedJsonModelSerializerMixin(serializers.ModelSerializer):

    def __init__(self, *args, **kargs):
        super().__init__(*args, **kargs)

        self.serializer_field_mapping[ModelValidatedJSONField] = ValidatedJSONField

    def build_standard_field(self, field_name, model_field):
        field_class, kwargs =  super().build_standard_field(field_name, model_field)

        if isinstance(model_field, ModelValidatedJSONField):
            kwargs['schema'] = model_field.schema
            kwargs['json_validator_cls'] = model_field.json_validator_cls

        return field_class, kwargs



try:
    from drf_spectacular.extensions import OpenApiSerializerFieldExtension
    from drf_spectacular.plumbing import build_basic_type
    from drf_spectacular.types import OpenApiTypes

    class ValidatedJsonFieldExtension(OpenApiSerializerFieldExtension):
        target_class = ValidatedJSONField

        def map_serializer_field(self, auto_schema, direction):
            #print("schema", self.target.schema)
            return self.target.schema or {"type":"object"}
except: pass

    
